@@ -2,12 +2,10 @@
 
    \brief ClamFS main file
 
-   $Id: clamfs.cxx,v 1.22 2008-12-06 13:27:30 burghardt Exp $
-
 *//*
 
    ClamFS - An user-space anti-virus protected file system
-   Copyright (C) 2007,2008 Krzysztof Burghardt.
+   Copyright (C) 2007-2019 Krzysztof Burghardt
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -35,8 +33,10 @@
     See the file COPYING.
 */
 
+#include "version.h"
 #include "config.h"
 
+#include <iostream>
 #include <fuse.h>
 #include <stdio.h>
 #include <string.h>
@@ -53,6 +53,7 @@
 #include "clamfs.hxx"
 #include "utils.hxx"
 
+using namespace std;
 using namespace boost;
 using namespace clamfs;
 
@@ -88,7 +89,19 @@ static inline const char* fixpath(const char* path)
 {
     char* fixed = new char[strlen(path)+2];
 
-    fchdir(savefd);
+    int res = fchdir(savefd);
+
+    if (res < 0)
+    {
+        char* username = getusername();
+        char* callername = getcallername();
+        rLog(Warn, "(%s:%d) (%s:%d) %s: fchdir() failed: %s",
+                callername, fuse_get_context()->pid, username, fuse_get_context()->uid,
+                path, strerror(errno));
+        free(username);
+        free(callername);
+    }
+
     strcpy(fixed,".");
     strcat(fixed,path);
 
@@ -126,7 +139,7 @@ static int clamfs_fgetattr(const char *path, struct stat *stbuf,
 
     (void) path;
 
-    res = fstat(fi->fh, stbuf);
+    res = fstat((int)fi->fh, stbuf);
     if (res == -1)
         return -errno;
 
@@ -159,7 +172,7 @@ static int clamfs_access(const char *path, int mask)
 */
 static int clamfs_readlink(const char *path, char *buf, size_t size)
 {
-    int res;
+    ssize_t res;
 
     const char* fpath = fixpath(path);
     res = readlink(fpath, buf, size - 1);
@@ -219,7 +232,7 @@ static int clamfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
         struct stat st;
         memset(&st, 0, sizeof(st));
         st.st_ino = de->d_ino;
-        st.st_mode = de->d_type << 12;
+        st.st_mode = (unsigned int)de->d_type << 12;
         if (filler(buf, de->d_name, &st, telldir(dp)))
             break;
     }
@@ -241,7 +254,7 @@ static int clamfs_releasedir(const char *path, struct fuse_file_info *fi)
 }
 
 /*!\brief FUSE mknod() callback
-   \param path file path 
+   \param path file path
    \param mode file permissions
    \param rdev major and minor numbers of device special file
    \returns 0 if mknod() returns without error on -errno otherwise
@@ -256,7 +269,10 @@ static int clamfs_mknod(const char *path, mode_t mode, dev_t rdev)
     else
         res = mknod(fpath, mode, rdev);
     if (res == -1)
+    {
+        delete[] fpath;
         return -errno;
+    }
     else
     res = lchown(fpath, fuse_get_context()->uid, fuse_get_context()->gid);
     delete[] fpath;
@@ -276,7 +292,10 @@ static int clamfs_mkdir(const char *path, mode_t mode)
     const char* fpath = fixpath(path);
     res = mkdir(fpath, mode);
     if (res == -1)
+    {
+        delete[] fpath;
         return -errno;
+    }
     else
     res = lchown(fpath, fuse_get_context()->uid, fuse_get_context()->gid);
     delete[] fpath;
@@ -327,15 +346,13 @@ static int clamfs_symlink(const char *from, const char *to)
 {
     int res;
 
-    const char* ffrom = fixpath(from);
     const char* fto = fixpath(to);
-    res = symlink(ffrom, fto);
+    res = symlink(from, fto);
+    delete[] fto;
     if (res == -1)
         return -errno;
     else
-    res = lchown(ffrom, fuse_get_context()->uid, fuse_get_context()->gid);
-    delete[] ffrom;
-    delete[] fto;
+    res = lchown(from, fuse_get_context()->uid, fuse_get_context()->gid);
 
     return 0;
 }
@@ -372,12 +389,15 @@ static int clamfs_link(const char *from, const char *to)
     const char* ffrom = fixpath(from);
     const char* fto = fixpath(to);
     res = link(ffrom, fto);
-    if (res == -1)
-        return -errno;
-    else
-    res = lchown(ffrom, fuse_get_context()->uid, fuse_get_context()->gid);
-    delete[] ffrom;
     delete[] fto;
+    if (res == -1)
+    {
+        delete[] ffrom;
+        return -errno;
+    }
+    else
+        res = lchown(ffrom, fuse_get_context()->uid, fuse_get_context()->gid);
+    delete[] ffrom;
 
     return 0;
 }
@@ -450,7 +470,7 @@ static int clamfs_ftruncate(const char *path, off_t size,
 
     (void) path;
 
-    res = ftruncate(fi->fh, size);
+    res = ftruncate((int)fi->fh, size);
     if (res == -1)
         return -errno;
 
@@ -497,7 +517,18 @@ static int clamfs_create(const char *path, mode_t mode, struct fuse_file_info *f
        res = lchown(fpath, fuse_get_context()->uid, fuse_get_context()->gid);
     delete[] fpath;
 
-    fi->fh = fd;
+    if (res < 0)
+    {
+        char* username = getusername();
+        char* callername = getcallername();
+        rLog(Warn, "(%s:%d) (%s:%d) %s: lchown() failed: %s",
+                callername, fuse_get_context()->pid, username, fuse_get_context()->uid,
+                path, strerror(errno));
+        free(username);
+        free(callername);
+    }
+
+    fi->fh = (unsigned long) fd;
     return 0;
 }
 
@@ -516,7 +547,7 @@ static inline int open_backend(const char *path, struct fuse_file_info *fi)
     if (fd == -1)
         return -errno;
 
-    fi->fh = fd;
+    fi->fh = (unsigned long) fd;
     return 0;
 }
 
@@ -563,7 +594,7 @@ static int clamfs_open(const char *path, struct fuse_file_info *fi)
                         {
                             INC_STAT_COUNTER(whitelistHit);
                             char* username = getusername();
-                            char* callername = getusername();
+                            char* callername = getcallername();
                             rLog(Warn, "(%s:%d) (%s:%d) %s: excluded from anti-virus scan because extension whitelisted ",
                                     callername, fuse_get_context()->pid, username, fuse_get_context()->uid, path);
                             free(username);
@@ -576,7 +607,7 @@ static int clamfs_open(const char *path, struct fuse_file_info *fi)
                             INC_STAT_COUNTER(blacklistHit);
                             file_is_blacklisted = true;
                             char* username = getusername();
-                            char* callername = getusername();
+                            char* callername = getcallername();
                             rLog(Warn, "(%s:%d) (%s:%d) %s: forced anti-virus scan because extension blacklisted ",
                                     callername, fuse_get_context()->pid, username, fuse_get_context()->uid, path);
                             free(username);
@@ -601,7 +632,7 @@ static int clamfs_open(const char *path, struct fuse_file_info *fi)
             if (file_stat.st_size > atoi(config["maximal-size"])) { /* file too big */
                 INC_STAT_COUNTER(tooBigFile);
                 char* username = getusername();
-                char* callername = getusername();
+                char* callername = getcallername();
                 rLog(Warn, "(%s:%d) (%s:%d) %s: excluded from anti-virus scan because file is too big (file size: %ld bytes)",
                         callername, fuse_get_context()->pid, username, fuse_get_context()->uid, path, (long int)file_stat.st_size);
                 free(username);
@@ -620,11 +651,11 @@ static int clamfs_open(const char *path, struct fuse_file_info *fi)
             ret = lstat(real_path.get(), &file_stat);
         if (!ret) { /* got file stat without error */
 
-            if (cache->has(file_stat.st_ino)) {
-                Poco::SharedPtr<CachedResult> ptr_val;
+            Poco::SharedPtr<CachedResult> ptr_val;
+
+            if ((ptr_val = cache->get(file_stat.st_ino))) {
                 INC_STAT_COUNTER(earlyCacheHit);
                 DEBUG("early cache hit for inode %ld", (unsigned long)file_stat.st_ino);
-                ptr_val = cache->get(file_stat.st_ino);
 
                 if (ptr_val->scanTimestamp == file_stat.st_mtime) {
                     INC_STAT_COUNTER(lateCacheHit);
@@ -650,8 +681,6 @@ static int clamfs_open(const char *path, struct fuse_file_info *fi)
                     /*
                      * Check for scan results and update cache
                      */
-                    Poco::SharedPtr<CachedResult> ptr_val;
-                    ptr_val = cache->get(file_stat.st_ino);
                     ptr_val->scanTimestamp = file_stat.st_mtime;
                     if (scan_result == 1) { /* virus found */
                         ptr_val->isClean = false;
@@ -740,14 +769,14 @@ static int clamfs_open(const char *path, struct fuse_file_info *fi)
 static int clamfs_read(const char *path, char *buf, size_t size, off_t offset,
                     struct fuse_file_info *fi)
 {
-    int res;
+    ssize_t res;
 
     (void) path;
-    res = pread(fi->fh, buf, size, offset);
+    res = pread((int)fi->fh, buf, size, offset);
     if (res == -1)
         res = -errno;
 
-    return res;
+    return (int)res;
 }
 
 /*!\brief FUSE write() callback
@@ -761,14 +790,14 @@ static int clamfs_read(const char *path, char *buf, size_t size, off_t offset,
 static int clamfs_write(const char *path, const char *buf, size_t size,
                      off_t offset, struct fuse_file_info *fi)
 {
-    int res;
+    ssize_t res;
 
     (void) path;
-    res = pwrite(fi->fh, buf, size, offset);
+    res = pwrite((int)fi->fh, buf, size, offset);
     if (res == -1)
         res = -errno;
 
-    return res;
+    return (int)res;
 }
 
 /*!\brief FUSE statfs() callback
@@ -797,7 +826,7 @@ static int clamfs_statfs(const char *path, struct statvfs *stbuf)
 static int clamfs_release(const char *path, struct fuse_file_info *fi)
 {
     (void) path;
-    close(fi->fh);
+    close((int)fi->fh);
 
     return 0;
 }
@@ -818,10 +847,10 @@ static int clamfs_fsync(const char *path, int isdatasync,
     (void) isdatasync;
 #else
     if (isdatasync)
-        res = fdatasync(fi->fh);
+        res = fdatasync((int)fi->fh);
     else
 #endif
-        res = fsync(fi->fh);
+        res = fsync((int)fi->fh);
     if (res == -1)
         return -errno;
 
@@ -859,13 +888,13 @@ static int clamfs_setxattr(const char *path, const char *name, const char *value
 static int clamfs_getxattr(const char *path, const char *name, char *value,
                     size_t size)
 {
-    int res;
+    ssize_t res;
     const char* fpath = fixpath(path);
     res = lgetxattr(fpath, name, value, size);
     delete[] fpath;
     if (res == -1)
         return -errno;
-    return res;
+    return (int)res;
 }
 
 /*!\brief FUSE listxattr() callback
@@ -876,13 +905,13 @@ static int clamfs_getxattr(const char *path, const char *name, char *value,
 */
 static int clamfs_listxattr(const char *path, char *list, size_t size)
 {
-    int res;
+    ssize_t res;
     const char* fpath = fixpath(path);
     res = llistxattr(fpath, list, size);
     delete[] fpath;
     if (res == -1)
         return -errno;
-    return res;
+    return (int)res;
 }
 
 /*!\brief FUSE removexattr() callback
@@ -907,6 +936,7 @@ static int clamfs_removexattr(const char *path, const char *name)
    \param argv arguments array
    \returns 0 on success, error code otherwise
 */
+int main(int argc, char *argv[]);
 int main(int argc, char *argv[])
 {
     int ret;
@@ -960,14 +990,17 @@ int main(int argc, char *argv[])
     RLogInit(argc, argv);
     RLogOpenStdio();
 
-    rLog(Info, "ClamFS v"VERSION);
-    rLog(Info, "Copyright (c) 2007,2008 Krzysztof Burghardt <krzysztof@burghardt.pl>");
-    rLog(Info, "http://clamfs.sourceforge.net/");
+    rLog(Info, "ClamFS v" VERSION " (git-" PACKAGE_VERSION_GIT_DESCRIBE ")");
+    rLog(Info, "Copyright (c) 2007-2019 Krzysztof Burghardt <krzysztof@burghardt.pl>");
+    rLog(Info, "https://github.com/burghardt/clamfs");
 
     /*
      * Check if we have one argument (other arguments are assumed RLog related)
      */
-    if (argc < 2) {
+    if ((argc < 2) ||
+        ((argc > 1) &&
+         ((strncmp(argv[1], "-h", strlen("-h")) == 0) ||
+          (strncmp(argv[1], "--help", strlen("--help")) == 0)))) {
         rLog(Warn, "ClamFS need to be invoked with one parameter - location of configuration file");
         rLog(Warn, "Example: %s /etc/clamfs/home.xml", argv[0]);
         return EXIT_FAILURE;
@@ -1021,15 +1054,33 @@ int main(int argc, char *argv[])
         fuse_argv[fuse_argc++] = strdup("-o");
         if ((config["nonempty"] != NULL) && /* public and nonempty */
             (strncmp(config["nonempty"], "yes", 3) == 0)) {
+#ifdef HAVE_SETXATTR_WITH_ACL_PATCH
+            fuse_argv[fuse_argc++] =
+                strdup("allow_other,default_permissions,acl,nonempty");
+#else
             fuse_argv[fuse_argc++] =
                 strdup("allow_other,default_permissions,nonempty");
+#endif
         } else { /* public without nonempty */
+#ifdef HAVE_SETXATTR_WITH_ACL_PATCH
+            fuse_argv[fuse_argc++] = strdup("allow_other,default_permissions,acl");
+#else
             fuse_argv[fuse_argc++] = strdup("allow_other,default_permissions");
+#endif
         }
     } else if ((config["nonempty"] != NULL) && /* private and nonempty */
         (strncmp(config["nonempty"], "yes", 3) == 0)) {
         fuse_argv[fuse_argc++] = strdup("-o");
+#ifdef HAVE_SETXATTR_WITH_ACL_PATCH
+        fuse_argv[fuse_argc++] = strdup("nonempty,default_permissions,acl");
+#else
         fuse_argv[fuse_argc++] = strdup("nonempty");
+#endif
+#ifdef HAVE_SETXATTR_WITH_ACL_PATCH
+    } else {
+        fuse_argv[fuse_argc++] = strdup("-o");
+        fuse_argv[fuse_argc++] = strdup("default_permissions,acl");
+#endif
     }
 
     if ((config["readonly"] != NULL) &&
@@ -1109,6 +1160,11 @@ int main(int argc, char *argv[])
         rLog(Info, "Statistics module disabled");
     }
 
+    if ((config["memory"] != NULL) &&
+        (strncmp(config["memory"], "yes", 3) == 0)) {
+        stats->enableMemoryStats();
+    }
+
     /*
      * Open configured logging target
      */
@@ -1121,7 +1177,7 @@ int main(int argc, char *argv[])
                 RLogOpenLogFile(config["filename"]);
                 RLogCloseStdio();
             } else {
-                rLog(Warn, "logging method 'file' choosen, but no log 'filename' given");
+                rLog(Warn, "logging method 'file' chosen, but no log 'filename' given");
                 return EXIT_FAILURE;
             }
         }
@@ -1137,7 +1193,7 @@ int main(int argc, char *argv[])
     /*
      * Start FUSE
      */
-    ret = fuse_main(fuse_argc, fuse_argv, &clamfs_oper);
+    ret = fuse_main(fuse_argc, fuse_argv, &clamfs_oper, NULL);
 
     for (unsigned int i = 0; i < FUSE_MAX_ARGS; ++i)
         if (fuse_argv[i])
@@ -1153,7 +1209,9 @@ int main(int argc, char *argv[])
     if (stats) {
         if ((config["atexit"] != NULL) &&
             (strncmp(config["atexit"], "yes", 3) == 0)) {
-            stats->dumpToLog();
+            stats->dumpFilesystemStatsToLog();
+            if (stats->memoryStats)
+                stats->dumpMemoryStatsToLog();
         }
 
         rLog(Info, "deleting stats");
